# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base form classes
"""

import json
import logging

import colander
import deform
from pyramid.renderers import render
from webhelpers2.html import HTML

from wuttaweb.util import get_form_data


log = logging.getLogger(__name__)


class FieldList(list):
    """
    Convenience wrapper for a form's field list.  This is a subclass
    of :class:`python:list`.

    You normally would not need to instantiate this yourself, but it
    is used under the hood for e.g. :attr:`Form.fields`.
    """

    def insert_before(self, field, newfield):
        """
        Insert a new field, before an existing field.

        :param field: String name for the existing field.

        :param newfield: String name for the new field, to be inserted
           just before the existing ``field``.
        """
        if field in self:
            i = self.index(field)
            self.insert(i, newfield)
        else:
            log.warning("field '%s' not found, will append new field: %s",
                        field, newfield)
            self.append(newfield)

    def insert_after(self, field, newfield):
        """
        Insert a new field, after an existing field.

        :param field: String name for the existing field.

        :param newfield: String name for the new field, to be inserted
           just after the existing ``field``.
        """
        if field in self:
            i = self.index(field)
            self.insert(i + 1, newfield)
        else:
            log.warning("field '%s' not found, will append new field: %s",
                        field, newfield)
            self.append(newfield)


class Form:
    """
    Base class for all forms.

    :param request: Reference to current :term:`request` object.

    :param fields: List of field names for the form.  This is
       optional; if not specified an attempt will be made to deduce
       the list automatically.  See also :attr:`fields`.

    :param schema: Colander-based schema object for the form.  This is
       optional; if not specified an attempt will be made to construct
       one automatically.  See also :meth:`get_schema()`.

    :param labels: Optional dict of default field labels.

    .. note::

       Some parameters are not explicitly described above.  However
       their corresponding attributes are described below.

    Form instances contain the following attributes:

    .. attribute:: fields

       :class:`FieldList` instance containing string field names for
       the form.  By default, fields will appear in the same order as
       they are in this list.

    .. attribute:: request

       Reference to current :term:`request` object.

    .. attribute:: action_url

       String URL to which the form should be submitted, if applicable.

    .. attribute:: vue_tagname

       String name for Vue component tag.  By default this is
       ``'wutta-form'``.  See also :meth:`render_vue_tag()`.

    .. attribute:: align_buttons_right

       Flag indicating whether the buttons (submit, cancel etc.)
       should be aligned to the right of the area below the form.  If
       not set, the buttons are left-aligned.

    .. attribute:: auto_disable_submit

       Flag indicating whether the submit button should be
       auto-disabled, whenever the form is submitted.

    .. attribute:: button_label_submit

       String label for the form submit button.  Default is ``"Save"``.

    .. attribute:: button_icon_submit

       String icon name for the form submit button.  Default is ``'save'``.

    .. attribute:: show_button_reset

       Flag indicating whether a Reset button should be shown.

    .. attribute:: validated

       If the :meth:`validate()` method was called, and it succeeded,
       this will be set to the validated data dict.

       Note that in all other cases, this attribute may not exist.
    """

    def __init__(
            self,
            request,
            fields=None,
            schema=None,
            labels={},
            action_url=None,
            vue_tagname='wutta-form',
            align_buttons_right=False,
            auto_disable_submit=True,
            button_label_submit="Save",
            button_icon_submit='save',
            show_button_reset=False,
    ):
        self.request = request
        self.schema = schema
        self.labels = labels or {}
        self.action_url = action_url
        self.vue_tagname = vue_tagname
        self.align_buttons_right = align_buttons_right
        self.auto_disable_submit = auto_disable_submit
        self.button_label_submit = button_label_submit
        self.button_icon_submit = button_icon_submit
        self.show_button_reset = show_button_reset

        self.config = self.request.wutta_config
        self.app = self.config.get_app()

        if fields is not None:
            self.set_fields(fields)
        elif self.schema:
            self.set_fields([f.name for f in self.schema])
        else:
            self.fields = None

    def __contains__(self, name):
        """
        Custom logic for the ``in`` operator, to allow easily checking
        if the form contains a given field::

           myform = Form()
           if 'somefield' in myform:
               print("my form has some field")
        """
        return bool(self.fields and name in self.fields)

    def __iter__(self):
        """
        Custom logic to allow iterating over form field names::

           myform = Form(fields=['foo', 'bar'])
           for fieldname in myform:
               print(fieldname)
        """
        return iter(self.fields)

    @property
    def vue_component(self):
        """
        String name for the Vue component, e.g. ``'WuttaForm'``.

        This is a generated value based on :attr:`vue_tagname`.
        """
        words = self.vue_tagname.split('-')
        return ''.join([word.capitalize() for word in words])

    def set_fields(self, fields):
        """
        Explicitly set the list of form fields.

        This will overwrite :attr:`fields` with a new
        :class:`FieldList` instance.

        :param fields: List of string field names.
        """
        self.fields = FieldList(fields)

    def set_label(self, key, label):
        """
        Set the label for given field name.

        See also :meth:`get_label()`.
        """
        self.labels[key] = label

        # update schema if necessary
        if self.schema and key in self.schema:
            self.schema[key].title = label

    def get_label(self, key):
        """
        Get the label for given field name.

        Note that this will always return a string, auto-generating
        the label if needed.

        See also :meth:`set_label()`.
        """
        return self.labels.get(key, self.app.make_title(key))

    def get_schema(self):
        """
        Return the :class:`colander:colander.Schema` object for the
        form, generating it automatically if necessary.
        """
        if not self.schema:
            raise NotImplementedError

        return self.schema

    def get_deform(self):
        """
        Return the :class:`deform:deform.Form` instance for the form,
        generating it automatically if necessary.
        """
        if not hasattr(self, 'deform_form'):
            schema = self.get_schema()
            form = deform.Form(schema)
            self.deform_form = form

        return self.deform_form

    def render_vue_tag(self, **kwargs):
        """
        Render the Vue component tag for the form.

        By default this simply returns:

        .. code-block:: html

           <wutta-form></wutta-form>

        The actual output will depend on various form attributes, in
        particular :attr:`vue_tagname`.
        """
        return HTML.tag(self.vue_tagname, **kwargs)

    def render_vue_template(
            self,
            template='/forms/vue_template.mako',
            **context):
        """
        Render the Vue template block for the form.

        This returns something like:

        .. code-block:: none

           <script type="text/x-template" id="wutta-form-template">
             <form>
               <!-- fields etc. -->
             </form>
           </script>

        .. todo::

           Why can't Sphinx render the above code block as 'html' ?

           It acts like it can't handle a ``<script>`` tag at all?

        Actual output will of course depend on form attributes, i.e.
        :attr:`vue_tagname` and :attr:`fields` list etc.

        :param template: Path to Mako template which is used to render
           the output.
        """
        context['form'] = self
        context.setdefault('form_attrs', {})
        context.setdefault('request', self.request)

        # auto disable button on submit
        if self.auto_disable_submit:
            context['form_attrs']['@submit'] = 'formSubmitting = true'

        output = render(template, context)
        return HTML.literal(output)

    def render_vue_field(self, fieldname):
        """
        Render the given field completely, i.e. ``<b-field>`` wrapper
        with label and containing a widget.

        Actual output will depend on the field attributes etc.
        Typical output might look like:

        .. code-block:: html

           <b-field label="Foo"
                    horizontal
                    type="is-danger"
                    message="something went wrong!">
              <!-- widget element(s) -->
           </b-field>
        """
        dform = self.get_deform()
        field = dform[fieldname]

        # render the field widget or whatever
        html = field.serialize()
        html = HTML.literal(html)

        # render field label
        label = self.get_label(fieldname)

        # b-field attrs
        attrs = {
            ':horizontal': 'true',
            'label': label,
        }

        # next we will build array of messages to display..some
        # fields always show a "helptext" msg, and some may have
        # validation errors..
        field_type = None
        messages = []

        # show errors if present
        errors = self.get_field_errors(fieldname)
        if errors:
            field_type = 'is-danger'
            messages.extend(errors)

        # ..okay now we can declare the field messages and type
        if field_type:
            attrs['type'] = field_type
        if messages:
            if len(messages) == 1:
                msg = messages[0]
                if msg.startswith('`') and msg.endswith('`'):
                    attrs[':message'] = msg
                else:
                    attrs['message'] = msg
            # TODO
            # else:
            #     # nb. must pass an array as JSON string
            #     attrs[':message'] = '[{}]'.format(', '.join([
            #         "'{}'".format(msg.replace("'", r"\'"))
            #         for msg in messages]))

        return HTML.tag('b-field', c=[html], **attrs)

    def get_field_errors(self, field):
        """
        Return a list of error messages for the given field.

        Not useful unless a call to :meth:`validate()` failed.
        """
        dform = self.get_deform()
        if field in dform:
            error = dform[field].errormsg
            if error:
                return [error]
        return []

    def get_vue_field_value(self, field):
        """
        This method returns a JSON string which will be assigned as
        the initial model value for the given field.  This JSON will
        be written as part of the overall response, to be interpreted
        on the client side.

        Again, this must return a *string* such as:

        * ``'null'``
        * ``'{"foo": "bar"}'``

        In practice this calls :meth:`jsonify_value()` to convert the
        ``field.cstruct`` value to string.
        """
        if isinstance(field, str):
            dform = self.get_deform()
            field = dform[field]

        return self.jsonify_value(field.cstruct)

    def jsonify_value(self, value):
        """
        Convert a Python value to JSON string.

        See also :meth:`get_vue_field_value()`.
        """
        if value is colander.null:
            return 'null'

        return json.dumps(value)

    def validate(self):
        """
        Try to validate the form.

        This should work whether request data was submitted as classic
        POST data, or as JSON body.

        If the form data is valid, this method returns the data dict.
        This data dict is also then available on the form object via
        the :attr:`validated` attribute.

        However if the data is not valid, ``False`` is returned, and
        there will be no :attr:`validated` attribute.  In that case
        you should inspect the form errors to learn/display what went
        wrong for the user's sake.  See also
        :meth:`get_field_errors()`.

        :returns: Data dict, or ``False``.
        """
        if hasattr(self, 'validated'):
            del self.validated

        if self.request.method != 'POST':
            return False

        dform = self.get_deform()
        controls = get_form_data(self.request).items()

        try:
            self.validated = dform.validate(controls)
        except deform.ValidationFailure:
            return False

        return self.validated



# -*- coding: utf-8; -*-

from wuttaweb.menus import MenuHandler


class NullMenuHandler(MenuHandler):
    """
    Dummy menu handler for testing.
    """
    def make_menus(self, request, **kwargs):
        return []



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import MagicMock

from pyramid import testing
from pyramid.httpexceptions import HTTPFound, HTTPForbidden

from wuttjamaican.conf import WuttaConfig
from wuttaweb.views import auth as mod
from wuttaweb.auth import WuttaSecurityPolicy
from wuttaweb.subscribers import new_request


class TestAuthView(TestCase):

    def setUp(self):
        self.config = WuttaConfig(defaults={
            'wutta.db.default.url': 'sqlite://',
        })

        self.request = testing.DummyRequest(wutta_config=self.config, user=None)
        self.pyramid_config = testing.setUp(request=self.request, settings={
            'wutta_config': self.config,
        })

        self.app = self.config.get_app()
        auth = self.app.get_auth_handler()
        model = self.app.model
        model.Base.metadata.create_all(bind=self.config.appdb_engine)
        self.session = self.app.make_session()
        self.user = model.User(username='barney')
        self.session.add(self.user)
        auth.set_user_password(self.user, 'testpass')
        self.session.commit()

        self.pyramid_config.set_security_policy(WuttaSecurityPolicy(db_session=self.session))
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.include('wuttaweb.views.common')

    def tearDown(self):
        testing.tearDown()

    def test_login(self):
        view = mod.AuthView(self.request)
        context = view.login()
        self.assertIn('form', context)

        # redirect if user already logged in
        self.request.user = self.user
        view = mod.AuthView(self.request)
        redirect = view.login(session=self.session)
        self.assertIsInstance(redirect, HTTPFound)

        # login fails w/ wrong password
        self.request.user = None
        self.request.method = 'POST'
        self.request.POST = {'username': 'barney', 'password': 'WRONG'}
        view = mod.AuthView(self.request)
        context = view.login(session=self.session)
        self.assertIn('form', context)

        # redirect if login succeeds
        self.request.method = 'POST'
        self.request.POST = {'username': 'barney', 'password': 'testpass'}
        view = mod.AuthView(self.request)
        redirect = view.login(session=self.session)
        self.assertIsInstance(redirect, HTTPFound)

    def test_logout(self):
        view = mod.AuthView(self.request)
        self.request.session.delete = MagicMock()
        redirect = view.logout()
        self.request.session.delete.assert_called_once_with()
        self.assertIsInstance(redirect, HTTPFound)

    def test_change_password(self):
        view = mod.AuthView(self.request)
        auth = self.app.get_auth_handler()

        # unauthenticated user is redirected
        redirect = view.change_password()
        self.assertIsInstance(redirect, HTTPFound)

        # now "login" the user, and set initial password
        self.request.user = self.user
        auth.set_user_password(self.user, 'foo')
        self.session.commit()

        # view should now return context w/ form
        context = view.change_password()
        self.assertIn('form', context)

        # submit valid form, ensure password is changed
        # (nb. this also would redirect user to home page)
        self.request.method = 'POST'
        self.request.POST = {
            'current_password': 'foo',
            # nb. new_password requires colander mapping structure
            '__start__': 'new_password:mapping',
            'new_password': 'bar',
            'new_password-confirm': 'bar',
            '__end__': 'new_password:mapping',
        }
        redirect = view.change_password()
        self.assertIsInstance(redirect, HTTPFound)
        self.session.commit()
        self.session.refresh(self.user)
        self.assertFalse(auth.check_user_password(self.user, 'foo'))
        self.assertTrue(auth.check_user_password(self.user, 'bar'))

        # at this point 'foo' is the password, now let's submit some
        # invalid forms and make sure we get back a context w/ form

        # first try empty data
        self.request.POST = {}
        context = view.change_password()
        self.assertIn('form', context)
        dform = context['form'].get_deform()
        self.assertEqual(dform['current_password'].errormsg, "Required")
        self.assertEqual(dform['new_password'].errormsg, "Required")

        # now try bad current password
        self.request.POST = {
            'current_password': 'blahblah',
            '__start__': 'new_password:mapping',
            'new_password': 'baz',
            'new_password-confirm': 'baz',
            '__end__': 'new_password:mapping',
        }
        context = view.change_password()
        self.assertIn('form', context)
        dform = context['form'].get_deform()
        self.assertEqual(dform['current_password'].errormsg, "Current password is incorrect.")

        # now try bad new password
        self.request.POST = {
            'current_password': 'bar',
            '__start__': 'new_password:mapping',
            'new_password': 'bar',
            'new_password-confirm': 'bar',
            '__end__': 'new_password:mapping',
        }
        context = view.change_password()
        self.assertIn('form', context)
        dform = context['form'].get_deform()
        self.assertEqual(dform['new_password'].errormsg, "New password must be different from old password.")

    def test_become_root(self):
        event = MagicMock(request=self.request)
        new_request(event)      # add request.get_referrer()
        view = mod.AuthView(self.request)

        # GET not allowed
        self.request.method = 'GET'
        self.assertRaises(HTTPForbidden, view.become_root)

        # non-admin users also not allowed
        self.request.method = 'POST'
        self.request.is_admin = False
        self.assertRaises(HTTPForbidden, view.become_root)

        # but admin users can become root
        self.request.is_admin = True
        self.assertNotIn('is_root', self.request.session)
        redirect = view.become_root()
        self.assertIsInstance(redirect, HTTPFound)
        self.assertTrue(self.request.session['is_root'])

    def test_stop_root(self):
        event = MagicMock(request=self.request)
        new_request(event)      # add request.get_referrer()
        view = mod.AuthView(self.request)

        # GET not allowed
        self.request.method = 'GET'
        self.assertRaises(HTTPForbidden, view.stop_root)

        # non-admin users also not allowed
        self.request.method = 'POST'
        self.request.is_admin = False
        self.assertRaises(HTTPForbidden, view.stop_root)

        # but admin users can stop being root
        # (nb. there is no check whether user is currently root)
        self.request.is_admin = True
        self.assertNotIn('is_root', self.request.session)
        redirect = view.stop_root()
        self.assertIsInstance(redirect, HTTPFound)
        self.assertFalse(self.request.session['is_root'])



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Web Handler

This defines the :term:`handler` for the web layer.
"""

from wuttjamaican.app import GenericHandler


class WebHandler(GenericHandler):
    """
    Base class and default implementation for the "web" :term:`handler`.

    This is responsible for determining the "menu handler" and
    (eventually) possibly other things.
    """

    def get_menu_handler(self, **kwargs):
        """
        Get the configured "menu" handler for the web app.

        Specify a custom handler in your config file like this:

        .. code-block:: ini

           [wutta.web]
           menus.handler_spec = poser.web.menus:PoserMenuHandler

        :returns: Instance of :class:`~wuttaweb.menus.MenuHandler`.
        """
        if not hasattr(self, 'menu_handler'):
            spec = self.config.get(f'{self.appname}.web.menus.handler_spec',
                                   default='wuttaweb.menus:MenuHandler')
            self.menu_handler = self.app.load_object(spec)(self.config)
        return self.menu_handler



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth Utility Logic
"""

import re

from pyramid.authentication import SessionAuthenticationHelper
from pyramid.request import RequestLocalCache
from pyramid.security import remember, forget

from wuttaweb.db import Session


def login_user(request, user):
    """
    Perform the steps necessary to "login" the given user.  This
    returns a ``headers`` dict which you should pass to the final
    redirect, like so::

       from pyramid.httpexceptions import HTTPFound

       headers = login_user(request, user)
       return HTTPFound(location='/', headers=headers)

    .. warning::

       This logic does not "authenticate" the user!  It assumes caller
       has already authenticated the user and they are safe to login.

    See also :func:`logout_user()`.
    """
    headers = remember(request, user.uuid)
    return headers


def logout_user(request):
    """
    Perform the logout action for the given request.  This returns a
    ``headers`` dict which you should pass to the final redirect, like
    so::

       from pyramid.httpexceptions import HTTPFound

       headers = logout_user(request)
       return HTTPFound(location='/', headers=headers)

    See also :func:`login_user()`.
    """
    request.session.delete()
    request.session.invalidate()
    headers = forget(request)
    return headers


class WuttaSecurityPolicy:
    """
    Pyramid :term:`security policy` for WuttaWeb.

    For more on the Pyramid details, see :doc:`pyramid:narr/security`.

    But the idea here is that you should be able to just use this,
    without thinking too hard::

       from pyramid.config import Configurator
       from wuttaweb.auth import WuttaSecurityPolicy

       pyramid_config = Configurator()
       pyramid_config.set_security_policy(WuttaSecurityPolicy())

    This security policy will then do the following:

    * use the request "web session" for auth storage (e.g. current
      ``user.uuid``)
    * check permissions as needed, by calling
      :meth:`~wuttjamaican:wuttjamaican.auth.AuthHandler.has_permission()`
      for current user

    :param db_session: Optional :term:`db session` to use, instead of
       :class:`wuttaweb.db.Session`.  Probably only useful for tests.
    """

    def __init__(self, db_session=None):
        self.session_helper = SessionAuthenticationHelper()
        self.identity_cache = RequestLocalCache(self.load_identity)
        self.db_session = db_session or Session()

    def load_identity(self, request):
        config = request.registry.settings['wutta_config']
        app = config.get_app()
        model = app.model

        # fetch user uuid from current session
        uuid = self.session_helper.authenticated_userid(request)
        if not uuid:
            return

        # fetch user object from db
        user = self.db_session.get(model.User, uuid)
        if not user:
            return

        return user

    def identity(self, request):
        return self.identity_cache.get_or_create(request)

    def authenticated_userid(self, request):
        user = self.identity(request)
        if user is not None:
            return user.uuid

    def remember(self, request, userid, **kw):
        return self.session_helper.remember(request, userid, **kw)

    def forget(self, request, **kw):
        return self.session_helper.forget(request, **kw)

    def permits(self, request, context, permission):

        # nb. root user can do anything
        if getattr(request, 'is_root', False):
            return True

        config = request.registry.settings['wutta_config']
        app = config.get_app()
        auth = app.get_auth_handler()
        user = self.identity(request)
        return auth.has_permission(self.db_session, user, permission)



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import MagicMock

import colander
import deform
from pyramid import testing

from wuttjamaican.conf import WuttaConfig
from wuttaweb.forms import base
from wuttaweb import helpers


class TestFieldList(TestCase):

    def test_insert_before(self):
        fields = base.FieldList(['f1', 'f2'])
        self.assertEqual(fields, ['f1', 'f2'])

        # typical
        fields.insert_before('f1', 'XXX')
        self.assertEqual(fields, ['XXX', 'f1', 'f2'])
        fields.insert_before('f2', 'YYY')
        self.assertEqual(fields, ['XXX', 'f1', 'YYY', 'f2'])

        # appends new field if reference field is invalid
        fields.insert_before('f3', 'ZZZ')
        self.assertEqual(fields, ['XXX', 'f1', 'YYY', 'f2', 'ZZZ'])

    def test_insert_after(self):
        fields = base.FieldList(['f1', 'f2'])
        self.assertEqual(fields, ['f1', 'f2'])

        # typical
        fields.insert_after('f1', 'XXX')
        self.assertEqual(fields, ['f1', 'XXX', 'f2'])
        fields.insert_after('XXX', 'YYY')
        self.assertEqual(fields, ['f1', 'XXX', 'YYY', 'f2'])

        # appends new field if reference field is invalid
        fields.insert_after('f3', 'ZZZ')
        self.assertEqual(fields, ['f1', 'XXX', 'YYY', 'f2', 'ZZZ'])


class TestForm(TestCase):

    def setUp(self):
        self.config = WuttaConfig(defaults={
            'wutta.web.menus.handler_spec': 'tests.utils:NullMenuHandler',
        })
        self.request = testing.DummyRequest(wutta_config=self.config, use_oruga=False)

        self.pyramid_config = testing.setUp(request=self.request, settings={
            'mako.directories': ['wuttaweb:templates'],
            'pyramid_deform.template_search_path': 'wuttaweb:templates/deform',
        })

    def tearDown(self):
        testing.tearDown()

    def make_form(self, request=None, **kwargs):
        return base.Form(request or self.request, **kwargs)

    def make_schema(self):
        schema = colander.Schema(children=[
            colander.SchemaNode(colander.String(),
                                name='foo'),
            colander.SchemaNode(colander.String(),
                                name='bar'),
        ])
        return schema

    def test_init_with_none(self):
        form = self.make_form()
        self.assertIsNone(form.fields)

    def test_init_with_fields(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.fields, ['foo', 'bar'])

    def test_init_with_schema(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        self.assertEqual(form.fields, ['foo', 'bar'])

    def test_vue_tagname(self):
        form = self.make_form()
        self.assertEqual(form.vue_tagname, 'wutta-form')

    def test_vue_component(self):
        form = self.make_form()
        self.assertEqual(form.vue_component, 'WuttaForm')

    def test_contains(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertIn('foo', form)
        self.assertNotIn('baz', form)

    def test_iter(self):
        form = self.make_form(fields=['foo', 'bar'])

        fields = list(iter(form))
        self.assertEqual(fields, ['foo', 'bar'])

        fields = []
        for field in form:
            fields.append(field)
        self.assertEqual(fields, ['foo', 'bar'])

    def test_set_fields(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.fields, ['foo', 'bar'])
        form.set_fields(['baz'])
        self.assertEqual(form.fields, ['baz'])

    def test_get_schema(self):
        form = self.make_form()
        self.assertIsNone(form.schema)

        # provided schema is returned
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        self.assertIs(form.schema, schema)
        self.assertIs(form.get_schema(), schema)

        # auto-generating schema not yet supported
        form = self.make_form(fields=['foo', 'bar'])
        self.assertIsNone(form.schema)
        self.assertRaises(NotImplementedError, form.get_schema)

    def test_get_deform(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        self.assertFalse(hasattr(form, 'deform_form'))
        dform = form.get_deform()
        self.assertIsInstance(dform, deform.Form)
        self.assertIs(form.deform_form, dform)

    def test_get_label(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.get_label('foo'), "Foo")
        form.set_label('foo', "Baz")
        self.assertEqual(form.get_label('foo'), "Baz")

    def test_set_label(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.get_label('foo'), "Foo")
        form.set_label('foo', "Baz")
        self.assertEqual(form.get_label('foo'), "Baz")

        # schema should be updated when setting label
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        form.set_label('foo', "Woohoo")
        self.assertEqual(form.get_label('foo'), "Woohoo")
        self.assertEqual(schema['foo'].title, "Woohoo")

    def test_render_vue_tag(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        html = form.render_vue_tag()
        self.assertEqual(html, '<wutta-form></wutta-form>')

    def test_render_vue_template(self):
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')

        # form button is disabled on @submit by default
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        html = form.render_vue_template()
        self.assertIn('<script type="text/x-template" id="wutta-form-template">', html)
        self.assertIn('@submit', html)

        # but not if form is configured otherwise
        form = self.make_form(schema=schema, auto_disable_submit=False)
        html = form.render_vue_template()
        self.assertIn('<script type="text/x-template" id="wutta-form-template">', html)
        self.assertNotIn('@submit', html)

    def test_render_vue_field(self):
        self.pyramid_config.include('pyramid_deform')
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        dform = form.get_deform()

        # typical
        html = form.render_vue_field('foo')
        self.assertIn('<b-field :horizontal="true" label="Foo">', html)
        self.assertIn('<b-input name="foo"', html)
        # nb. no error message
        self.assertNotIn('message', html)

        # with single "static" error
        dform['foo'].error = MagicMock(msg="something is wrong")
        html = form.render_vue_field('foo')
        self.assertIn(' message="something is wrong"', html)

        # with single "dynamic" error
        dform['foo'].error = MagicMock(msg="`something is wrong`")
        html = form.render_vue_field('foo')
        self.assertIn(':message="`something is wrong`"', html)

    def test_get_field_errors(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        dform = form.get_deform()

        # no error
        errors = form.get_field_errors('foo')
        self.assertEqual(len(errors), 0)

        # simple error
        dform['foo'].error = MagicMock(msg="something is wrong")
        errors = form.get_field_errors('foo')
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "something is wrong")

    def test_get_vue_field_value(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)

        # null field value
        value = form.get_vue_field_value('foo')
        self.assertEqual(value, 'null')

        # non-default / explicit value
        # TODO: surely need a different approach to set value
        dform = form.get_deform()
        dform['foo'].cstruct = 'blarg'
        value = form.get_vue_field_value('foo')
        self.assertEqual(value, '"blarg"')

    def test_jsonify_value(self):
        form = self.make_form()

        # null field value
        value = form.jsonify_value(colander.null)
        self.assertEqual(value, 'null')
        value = form.jsonify_value(None)
        self.assertEqual(value, 'null')

        # string value
        value = form.jsonify_value('blarg')
        self.assertEqual(value, '"blarg"')

    def test_validate(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        self.assertFalse(hasattr(form, 'validated'))

        # will not validate unless request is POST
        self.request.POST = {'foo': 'blarg', 'bar': 'baz'}
        self.request.method = 'GET'
        self.assertFalse(form.validate())
        self.request.method = 'POST'
        data = form.validate()
        self.assertEqual(data, {'foo': 'blarg', 'bar': 'baz'})

        # validating a second type updates form.validated
        self.request.POST = {'foo': 'BLARG', 'bar': 'BAZ'}
        data = form.validate()
        self.assertEqual(data, {'foo': 'BLARG', 'bar': 'BAZ'})
        self.assertIs(form.validated, data)

        # bad data does not validate
        self.request.POST = {'foo': 42, 'bar': None}
        self.assertFalse(form.validate())
        dform = form.get_deform()
        self.assertEqual(len(dform.error.children), 2)
        self.assertEqual(dform['foo'].errormsg, "Pstruct is not a string")



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base Logic for Master Views
"""

from pyramid.renderers import render_to_response

from wuttaweb.views import View
from wuttaweb.util import get_form_data
from wuttaweb.db import Session


class MasterView(View):
    """
    Base class for "master" views.

    Master views typically map to a table in a DB, though not always.
    They essentially are a set of CRUD views for a certain type of
    data record.

    Many attributes may be overridden in subclass.  For instance to
    define :attr:`model_class`::

       from wuttaweb.views import MasterView
       from wuttjamaican.db.model import Person

       class MyPersonView(MasterView):
           model_class = Person

       def includeme(config):
           MyPersonView.defaults(config)

    .. note::

       Many of these attributes will only exist if they have been
       explicitly defined in a subclass.  There are corresponding
       ``get_xxx()`` methods which should be used instead of accessing
       these attributes directly.

    .. attribute:: model_class

       Optional reference to a data model class.  While not strictly
       required, most views will set this to a SQLAlchemy mapped
       class,
       e.g. :class:`wuttjamaican:wuttjamaican.db.model.auth.User`.

       Code should not access this directly but instead call
       :meth:`get_model_class()`.

    .. attribute:: model_name

       Optional override for the view's data model name,
       e.g. ``'WuttaWidget'``.

       Code should not access this directly but instead call
       :meth:`get_model_name()`.

    .. attribute:: model_name_normalized

       Optional override for the view's "normalized" data model name,
       e.g. ``'wutta_widget'``.

       Code should not access this directly but instead call
       :meth:`get_model_name_normalized()`.

    .. attribute:: model_title

       Optional override for the view's "humanized" (singular) model
       title, e.g. ``"Wutta Widget"``.

       Code should not access this directly but instead call
       :meth:`get_model_title()`.

    .. attribute:: model_title_plural

       Optional override for the view's "humanized" (plural) model
       title, e.g. ``"Wutta Widgets"``.

       Code should not access this directly but instead call
       :meth:`get_model_title_plural()`.

    .. attribute:: config_title

       Optional override for the view's "config" title, e.g. ``"Wutta
       Widgets"`` (to be displayed as **Configure Wutta Widgets**).

       Code should not access this directly but instead call
       :meth:`get_config_title()`.

    .. attribute:: route_prefix

       Optional override for the view's route prefix,
       e.g. ``'wutta_widgets'``.

       Code should not access this directly but instead call
       :meth:`get_route_prefix()`.

    .. attribute:: url_prefix

       Optional override for the view's URL prefix,
       e.g. ``'/widgets'``.

       Code should not access this directly but instead call
       :meth:`get_url_prefix()`.

    .. attribute:: template_prefix

       Optional override for the view's template prefix,
       e.g. ``'/widgets'``.

       Code should not access this directly but instead call
       :meth:`get_template_prefix()`.

    .. attribute:: listable

       Boolean indicating whether the view model supports "listing" -
       i.e. it should have an :meth:`index()` view.  Default value is
       ``True``.

    .. attribute:: configurable

       Boolean indicating whether the master view supports
       "configuring" - i.e. it should have a :meth:`configure()` view.
       Default value is ``False``.
    """

    ##############################
    # attributes
    ##############################

    # features
    listable = True
    configurable = False

    # current action
    configuring = False

    ##############################
    # index methods
    ##############################

    def index(self):
        """
        View to "list" (filter/browse) the model data.

        This is the "default" view for the model and is what user sees
        when visiting the "root" path under the :attr:`url_prefix`,
        e.g. ``/widgets/``.

        By default, this view is included only if :attr:`listable` is
        true.
        """
        context = {
            'index_url': None,  # avoid title link since this *is* the index
        }
        return self.render_to_response('index', context)

    ##############################
    # configure methods
    ##############################

    def configure(self):
        """
        View for configuring aspects of the app which are pertinent to
        this master view and/or model.

        By default, this view is included only if :attr:`configurable`
        is true.  It usually maps to a URL like ``/widgets/configure``.

        The expected workflow is as follows:

        * user navigates to Configure page
        * user modifies settings and clicks Save
        * this view then *deletes* all "known" settings
        * then it saves user-submitted settings

        That is unless ``remove_settings`` is requested, in which case
        settings are deleted but then none are saved.  The "known"
        settings by default include only the "simple" settings.

        As a general rule, a particular setting should be configurable
        by (at most) one master view.  Some settings may never be
        exposed at all.  But when exposing a setting, careful thought
        should be given to where it logically/best belongs.

        Some settings are "simple" and a master view subclass need
        only provide their basic definitions via
        :meth:`configure_get_simple_settings()`.  If complex settings
        are needed, subclass must override one or more other methods
        to achieve the aim(s).

        See also related methods, used by this one:

        * :meth:`configure_get_simple_settings()`
        * :meth:`configure_get_context()`
        * :meth:`configure_gather_settings()`
        * :meth:`configure_remove_settings()`
        * :meth:`configure_save_settings()`
        """
        self.configuring = True
        config_title = self.get_config_title()

        # was form submitted?
        if self.request.method == 'POST':

            # maybe just remove settings
            if self.request.POST.get('remove_settings'):
                self.configure_remove_settings()
                self.request.session.flash(f"All settings for {config_title} have been removed.",
                                           'warning')

                # reload configure page
                return self.redirect(self.request.current_route_url())

            # gather/save settings
            data = get_form_data(self.request)
            settings = self.configure_gather_settings(data)
            self.configure_remove_settings()
            self.configure_save_settings(settings)
            self.request.session.flash("Settings have been saved.")

            # reload configure page
            return self.redirect(self.request.current_route_url())

        # render configure page
        context = self.configure_get_context()
        return self.render_to_response('configure', context)

    def configure_get_context(
            self,
            simple_settings=None,
    ):
        """
        Returns the full context dict, for rendering the
        :meth:`configure()` page template.

        Default context will include ``simple_settings`` (normalized
        to just name/value).

        You may need to override this method, to add additional
        "complex" settings etc.

        :param simple_settings: Optional list of simple settings, if
           already initialized.  Otherwise it is retrieved via
           :meth:`configure_get_simple_settings()`.

        :returns: Context dict for the page template.
        """
        context = {}

        # simple settings
        if simple_settings is None:
            simple_settings = self.configure_get_simple_settings()
        if simple_settings:

            # we got some, so "normalize" each definition to name/value
            normalized = {}
            for simple in simple_settings:

                # name
                name = simple['name']

                # value
                if 'value' in simple:
                    value = simple['value']
                elif simple.get('type') is bool:
                    value = self.config.get_bool(name, default=simple.get('default', False))
                else:
                    value = self.config.get(name)

                normalized[name] = value

            # add to template context
            context['simple_settings'] = normalized

        return context

    def configure_get_simple_settings(self):
        """
        This should return a list of "simple" setting definitions for
        the :meth:`configure()` view, which can be handled in a more
        automatic way.  (This is as opposed to some settings which are
        more complex and must be handled manually; those should not be
        part of this method's return value.)

        Basically a "simple" setting is one which can be represented
        by a single field/widget on the Configure page.

        The setting definitions returned must each be a dict of
        "attributes" for the setting.  For instance a *very* simple
        setting might be::

           {'name': 'wutta.app_title'}

        The ``name`` is required, everything else is optional.  Here
        is a more complete example::

           {
               'name': 'wutta.production',
               'type': bool,
               'default': False,
               'save_if_empty': False,
           }

        Note that if specified, the ``default`` should be of the same
        data type as defined for the setting (``bool`` in the above
        example).  The default ``type`` is ``str``.

        Normally if a setting's value is effectively null, the setting
        is removed instead of keeping it in the DB.  This behavior can
        be changed per-setting via the ``save_if_empty`` flag.

        :returns: List of setting definition dicts as described above.
           Note that their order does not matter since the template
           must explicitly define field layout etc.
        """

    def configure_gather_settings(
            self,
            data,
            simple_settings=None,
    ):
        """
        Collect the full set of "normalized" settings from user
        request, so that :meth:`configure()` can save them.

        Settings are gathered from the given request (e.g. POST)
        ``data``, but also taking into account what we know based on
        the simple setting definitions.

        Subclass may need to override this method if complex settings
        are required.

        :param data: Form data submitted via POST request.

        :param simple_settings: Optional list of simple settings, if
           already initialized.  Otherwise it is retrieved via
           :meth:`configure_get_simple_settings()`.

        This method must return a list of normalized settings, similar
        in spirit to the definition syntax used in
        :meth:`configure_get_simple_settings()`.  However the format
        returned here is minimal and contains just name/value::

           {
               'name': 'wutta.app_title',
               'value': 'Wutta Wutta',
           }

        Note that the ``value`` will always be a string.

        Also note, whereas it's possible ``data`` will not contain all
        known settings, the return value *should* (potentially)
        contain all of them.

        The one exception is when a simple setting has null value, by
        default it will not be included in the result (hence, not
        saved to DB) unless the setting definition has the
        ``save_if_empty`` flag set.
        """
        settings = []

        # simple settings
        if simple_settings is None:
            simple_settings = self.configure_get_simple_settings()
        if simple_settings:

            # we got some, so "normalize" each definition to name/value
            for simple in simple_settings:
                name = simple['name']

                if name in data:
                    value = data[name]
                else:
                    value = simple.get('default')

                if simple.get('type') is bool:
                    value = str(bool(value)).lower()
                elif simple.get('type') is int:
                    value = str(int(value or '0'))
                elif value is None:
                    value = ''
                else:
                    value = str(value)

                # only want to save this setting if we received a
                # value, or if empty values are okay to save
                if value or simple.get('save_if_empty'):
                    settings.append({'name': name,
                                     'value': value})

        return settings

    def configure_remove_settings(
            self,
            simple_settings=None,
    ):
        """
        Remove all "known" settings from the DB; this is called by
        :meth:`configure()`.

        The point of this method is to ensure *all* "known" settings
        which are managed by this master view, are purged from the DB.

        The default logic can handle this automatically for simple
        settings; subclass must override for any complex settings.

        :param simple_settings: Optional list of simple settings, if
           already initialized.  Otherwise it is retrieved via
           :meth:`configure_get_simple_settings()`.
        """
        names = []

        # simple settings
        if simple_settings is None:
            simple_settings = self.configure_get_simple_settings()
        if simple_settings:
            names.extend([simple['name']
                          for simple in simple_settings])

        if names:
            # nb. must avoid self.Session here in case that does not
            # point to our primary app DB
            session = Session()
            for name in names:
                self.app.delete_setting(session, name)

    def configure_save_settings(self, settings):
        """
        Save the given settings to the DB; this is called by
        :meth:`configure()`.

        This method expected a list of name/value dicts and will
        simply save each to the DB, with no "conversion" logic.

        :param settings: List of normalized setting definitions, as
           returned by :meth:`configure_gather_settings()`.
        """
        # app = self.get_rattail_app()

        # nb. must avoid self.Session here in case that does not point
        # to our primary app DB
        session = Session()
        for setting in settings:
            self.app.save_setting(session, setting['name'], setting['value'],
                                  force_create=True)

    ##############################
    # support methods
    ##############################

    def get_index_title(self):
        """
        Returns the main index title for the master view.

        By default this returns the value from
        :meth:`get_model_title_plural()`.  Subclass may override as
        needed.
        """
        return self.get_model_title_plural()

    def get_index_url(self, **kwargs):
        """
        Returns the URL for master's :meth:`index()` view.

        NB. this returns ``None`` if :attr:`listable` is false.
        """
        if self.listable:
            route_prefix = self.get_route_prefix()
            return self.request.route_url(route_prefix, **kwargs)

    def render_to_response(self, template, context):
        """
        Locate and render an appropriate template, with the given
        context, and return a :term:`response`.

        The specified ``template`` should be only the "base name" for
        the template - e.g.  ``'index'`` or ``'edit'``.  This method
        will then try to locate a suitable template file, based on
        values from :meth:`get_template_prefix()` and
        :meth:`get_fallback_templates()`.

        In practice this *usually* means two different template paths
        will be attempted, e.g. if ``template`` is ``'edit'`` and
        :attr:`template_prefix` is ``'/widgets'``:

        * ``/widgets/edit.mako``
        * ``/master/edit.mako``

        The first template found to exist will be used for rendering.
        It then calls
        :func:`pyramid:pyramid.renderers.render_to_response()` and
        returns the result.

        :param template: Base name for the template.

        :param context: Data dict to be used as template context.

        :returns: Response object containing the rendered template.
        """
        defaults = {
            'master': self,
            'route_prefix': self.get_route_prefix(),
            'index_title': self.get_index_title(),
            'index_url': self.get_index_url(),
            'config_title': self.get_config_title(),
        }

        # merge defaults + caller-provided context
        defaults.update(context)
        context = defaults

        # first try the template path most specific to this view
        template_prefix = self.get_template_prefix()
        mako_path = f'{template_prefix}/{template}.mako'
        try:
            return render_to_response(mako_path, context, request=self.request)
        except IOError:

            # failing that, try one or more fallback templates
            for fallback in self.get_fallback_templates(template):
                try:
                    return render_to_response(fallback, context, request=self.request)
                except IOError:
                    pass

            # if we made it all the way here, then we found no
            # templates at all, in which case re-attempt the first and
            # let that error raise on up
            return render_to_response(mako_path, context, request=self.request)

    def get_fallback_templates(self, template):
        """
        Returns a list of "fallback" template paths which may be
        attempted for rendering a view.  This is used within
        :meth:`render_to_response()` if the "first guess" template
        file was not found.

        :param template: Base name for a template (without prefix), e.g.
           ``'custom'``.

        :returns: List of full template paths to be tried, based on
           the specified template.  For instance if ``template`` is
           ``'custom'`` this will (by default) return::

              ['/master/custom.mako']
        """
        return [f'/master/{template}.mako']

    ##############################
    # class methods
    ##############################

    @classmethod
    def get_model_class(cls):
        """
        Returns the model class for the view (if defined).

        A model class will *usually* be a SQLAlchemy mapped class,
        e.g. :class:`wuttjamaican:wuttjamaican.db.model.base.Person`.

        There is no default value here, but a subclass may override by
        assigning :attr:`model_class`.

        Note that the model class is not *required* - however if you
        do not set the :attr:`model_class`, then you *must* set the
        :attr:`model_name`.
        """
        if hasattr(cls, 'model_class'):
            return cls.model_class

    @classmethod
    def get_model_name(cls):
        """
        Returns the model name for the view.

        A model name should generally be in the format of a Python
        class name, e.g. ``'WuttaWidget'``.  (Note this is
        *singular*, not plural.)

        The default logic will call :meth:`get_model_class()` and
        return that class name as-is.  A subclass may override by
        assigning :attr:`model_name`.
        """
        if hasattr(cls, 'model_name'):
            return cls.model_name

        return cls.get_model_class().__name__

    @classmethod
    def get_model_name_normalized(cls):
        """
        Returns the "normalized" model name for the view.

        A normalized model name should generally be in the format of a
        Python variable name, e.g. ``'wutta_widget'``.  (Note this is
        *singular*, not plural.)

        The default logic will call :meth:`get_model_name()` and
        simply lower-case the result.  A subclass may override by
        assigning :attr:`model_name_normalized`.
        """
        if hasattr(cls, 'model_name_normalized'):
            return cls.model_name_normalized

        return cls.get_model_name().lower()

    @classmethod
    def get_model_title(cls):
        """
        Returns the "humanized" (singular) model title for the view.

        The model title will be displayed to the user, so should have
        proper grammar and capitalization, e.g. ``"Wutta Widget"``.
        (Note this is *singular*, not plural.)

        The default logic will call :meth:`get_model_name()` and use
        the result as-is.  A subclass may override by assigning
        :attr:`model_title`.
        """
        if hasattr(cls, 'model_title'):
            return cls.model_title

        return cls.get_model_name()

    @classmethod
    def get_model_title_plural(cls):
        """
        Returns the "humanized" (plural) model title for the view.

        The model title will be displayed to the user, so should have
        proper grammar and capitalization, e.g. ``"Wutta Widgets"``.
        (Note this is *plural*, not singular.)

        The default logic will call :meth:`get_model_title()` and
        simply add a ``'s'`` to the end.  A subclass may override by
        assigning :attr:`model_title_plural`.
        """
        if hasattr(cls, 'model_title_plural'):
            return cls.model_title_plural

        model_title = cls.get_model_title()
        return f"{model_title}s"

    @classmethod
    def get_route_prefix(cls):
        """
        Returns the "route prefix" for the master view.  This prefix
        is used for all named routes defined by the view class.

        For instance if route prefix is ``'widgets'`` then a view
        might have these routes:

        * ``'widgets'``
        * ``'widgets.create'``
        * ``'widgets.edit'``
        * ``'widgets.delete'``

        The default logic will call
        :meth:`get_model_name_normalized()` and simply add an ``'s'``
        to the end, making it plural.  A subclass may override by
        assigning :attr:`route_prefix`.
        """
        if hasattr(cls, 'route_prefix'):
            return cls.route_prefix

        model_name = cls.get_model_name_normalized()
        return f'{model_name}s'

    @classmethod
    def get_url_prefix(cls):
        """
        Returns the "URL prefix" for the master view.  This prefix is
        used for all URLs defined by the view class.

        Using the same example as in :meth:`get_route_prefix()`, the
        URL prefix would be ``'/widgets'`` and the view would have
        defined routes for these URLs:

        * ``/widgets/``
        * ``/widgets/new``
        * ``/widgets/XXX/edit``
        * ``/widgets/XXX/delete``

        The default logic will call :meth:`get_route_prefix()` and
        simply add a ``'/'`` to the beginning.  A subclass may
        override by assigning :attr:`url_prefix`.
        """
        if hasattr(cls, 'url_prefix'):
            return cls.url_prefix

        route_prefix = cls.get_route_prefix()
        return f'/{route_prefix}'

    @classmethod
    def get_template_prefix(cls):
        """
        Returns the "template prefix" for the master view.  This
        prefix is used to guess which template path to render for a
        given view.

        Using the same example as in :meth:`get_url_prefix()`, the
        template prefix would also be ``'/widgets'`` and the templates
        assumed for those routes would be:

        * ``/widgets/index.mako``
        * ``/widgets/create.mako``
        * ``/widgets/edit.mako``
        * ``/widgets/delete.mako``

        The default logic will call :meth:`get_url_prefix()` and
        return that value as-is.  A subclass may override by assigning
        :attr:`template_prefix`.
        """
        if hasattr(cls, 'template_prefix'):
            return cls.template_prefix

        return cls.get_url_prefix()

    @classmethod
    def get_config_title(cls):
        """
        Returns the "config title" for the view/model.

        The config title is used for page title in the
        :meth:`configure()` view, as well as links to it.  It is
        usually plural, e.g. ``"Wutta Widgets"`` in which case that
        winds up being displayed in the web app as: **Configure Wutta
        Widgets**

        The default logic will call :meth:`get_model_title_plural()`
        and return that as-is.  A subclass may override by assigning
        :attr:`config_title`.
        """
        if hasattr(cls, 'config_title'):
            return cls.config_title

        return cls.get_model_title_plural()

    ##############################
    # configuration
    ##############################

    @classmethod
    def defaults(cls, config):
        """
        Provide default Pyramid configuration for a master view.

        This is generally called from within the module's
        ``includeme()`` function, e.g.::

           from wuttaweb.views import MasterView

           class WidgetView(MasterView):
               model_name = 'Widget'

           def includeme(config):
               WidgetView.defaults(config)

        :param config: Reference to the app's
           :class:`pyramid:pyramid.config.Configurator` instance.
        """
        cls._defaults(config)

    @classmethod
    def _defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()

        # index
        if cls.listable:
            config.add_route(route_prefix, f'{url_prefix}/')
            config.add_view(cls, attr='index',
                            route_name=route_prefix)

        # configure
        if cls.configurable:
            config.add_route(f'{route_prefix}.configure',
                             f'{url_prefix}/configure')
            config.add_view(cls, attr='configure',
                            route_name=f'{route_prefix}.configure')



# -*- coding: utf-8; -*-

from tests.views.utils import WebTestCase

from wuttaweb.views import settings


class TestAppInfoView(WebTestCase):

    def test_index(self):
        # sanity/coverage check
        view = settings.AppInfoView(self.request)
        response = view.index()

    def test_configure_get_simple_settings(self):
        # sanity/coverage check
        view = settings.AppInfoView(self.request)
        simple = view.configure_get_simple_settings()

    def test_configure_get_context(self):
        # sanity/coverage check
        view = settings.AppInfoView(self.request)
        context = view.configure_get_context()



# -*- coding: utf-8; -*-

import functools
from unittest import TestCase
from unittest.mock import MagicMock, patch

from pyramid import testing
from pyramid.response import Response
from pyramid.httpexceptions import HTTPFound

from wuttjamaican.conf import WuttaConfig
from wuttaweb.views import master
from wuttaweb.subscribers import new_request_set_user

from tests.views.utils import WebTestCase


class TestMasterView(WebTestCase):

    def test_defaults(self):
        master.MasterView.model_name = 'Widget'
        # TODO: should inspect pyramid routes after this, to be certain
        master.MasterView.defaults(self.pyramid_config)
        del master.MasterView.model_name

    ##############################
    # class methods
    ##############################

    def test_get_model_class(self):
        
        # no model class by default
        self.assertIsNone(master.MasterView.get_model_class())

        # subclass may specify
        MyModel = MagicMock()
        master.MasterView.model_class = MyModel
        self.assertIs(master.MasterView.get_model_class(), MyModel)
        del master.MasterView.model_class

    def test_get_model_name(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_model_name)

        # subclass may specify model name
        master.MasterView.model_name = 'Widget'
        self.assertEqual(master.MasterView.get_model_name(), 'Widget')
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Blaster')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_model_name(), 'Blaster')
        del master.MasterView.model_class

    def test_get_model_name_normalized(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_model_name_normalized)

        # subclass may specify *normalized* model name
        master.MasterView.model_name_normalized = 'widget'
        self.assertEqual(master.MasterView.get_model_name_normalized(), 'widget')
        del master.MasterView.model_name_normalized

        # or it may specify *standard* model name
        master.MasterView.model_name = 'Blaster'
        self.assertEqual(master.MasterView.get_model_name_normalized(), 'blaster')
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_model_name_normalized(), 'dinosaur')
        del master.MasterView.model_class

    def test_get_model_title(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_model_title)

        # subclass may specify  model title
        master.MasterView.model_title = 'Wutta Widget'
        self.assertEqual(master.MasterView.get_model_title(), "Wutta Widget")
        del master.MasterView.model_title

        # or it may specify model name
        master.MasterView.model_name = 'Blaster'
        self.assertEqual(master.MasterView.get_model_title(), "Blaster")
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_model_title(), "Dinosaur")
        del master.MasterView.model_class

    def test_get_model_title_plural(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_model_title_plural)

        # subclass may specify *plural* model title
        master.MasterView.model_title_plural = 'People'
        self.assertEqual(master.MasterView.get_model_title_plural(), "People")
        del master.MasterView.model_title_plural

        # or it may specify *singular* model title
        master.MasterView.model_title = 'Wutta Widget'
        self.assertEqual(master.MasterView.get_model_title_plural(), "Wutta Widgets")
        del master.MasterView.model_title

        # or it may specify model name
        master.MasterView.model_name = 'Blaster'
        self.assertEqual(master.MasterView.get_model_title_plural(), "Blasters")
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_model_title_plural(), "Dinosaurs")
        del master.MasterView.model_class

    def test_get_route_prefix(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_route_prefix)

        # subclass may specify route prefix
        master.MasterView.route_prefix = 'widgets'
        self.assertEqual(master.MasterView.get_route_prefix(), 'widgets')
        del master.MasterView.route_prefix

        # subclass may specify *normalized* model name
        master.MasterView.model_name_normalized = 'blaster'
        self.assertEqual(master.MasterView.get_route_prefix(), 'blasters')
        del master.MasterView.model_name_normalized

        # or it may specify *standard* model name
        master.MasterView.model_name = 'Dinosaur'
        self.assertEqual(master.MasterView.get_route_prefix(), 'dinosaurs')
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Truck')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_route_prefix(), 'trucks')
        del master.MasterView.model_class

    def test_get_url_prefix(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_url_prefix)

        # subclass may specify url prefix
        master.MasterView.url_prefix = '/widgets'
        self.assertEqual(master.MasterView.get_url_prefix(), '/widgets')
        del master.MasterView.url_prefix

        # or it may specify route prefix
        master.MasterView.route_prefix = 'trucks'
        self.assertEqual(master.MasterView.get_url_prefix(), '/trucks')
        del master.MasterView.route_prefix

        # or it may specify *normalized* model name
        master.MasterView.model_name_normalized = 'blaster'
        self.assertEqual(master.MasterView.get_url_prefix(), '/blasters')
        del master.MasterView.model_name_normalized

        # or it may specify *standard* model name
        master.MasterView.model_name = 'Dinosaur'
        self.assertEqual(master.MasterView.get_url_prefix(), '/dinosaurs')
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Machine')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_url_prefix(), '/machines')
        del master.MasterView.model_class

    def test_get_template_prefix(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_template_prefix)

        # subclass may specify template prefix
        master.MasterView.template_prefix = '/widgets'
        self.assertEqual(master.MasterView.get_template_prefix(), '/widgets')
        del master.MasterView.template_prefix

        # or it may specify url prefix
        master.MasterView.url_prefix = '/trees'
        self.assertEqual(master.MasterView.get_template_prefix(), '/trees')
        del master.MasterView.url_prefix

        # or it may specify route prefix
        master.MasterView.route_prefix = 'trucks'
        self.assertEqual(master.MasterView.get_template_prefix(), '/trucks')
        del master.MasterView.route_prefix

        # or it may specify *normalized* model name
        master.MasterView.model_name_normalized = 'blaster'
        self.assertEqual(master.MasterView.get_template_prefix(), '/blasters')
        del master.MasterView.model_name_normalized

        # or it may specify *standard* model name
        master.MasterView.model_name = 'Dinosaur'
        self.assertEqual(master.MasterView.get_template_prefix(), '/dinosaurs')
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Machine')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_template_prefix(), '/machines')
        del master.MasterView.model_class

    def test_get_config_title(self):

        # error by default (since no model class)
        self.assertRaises(AttributeError, master.MasterView.get_config_title)

        # subclass may specify config title
        master.MasterView.config_title = 'Widgets'
        self.assertEqual(master.MasterView.get_config_title(), "Widgets")
        del master.MasterView.config_title

        # subclass may specify *plural* model title
        master.MasterView.model_title_plural = 'People'
        self.assertEqual(master.MasterView.get_config_title(), "People")
        del master.MasterView.model_title_plural

        # or it may specify *singular* model title
        master.MasterView.model_title = 'Wutta Widget'
        self.assertEqual(master.MasterView.get_config_title(), "Wutta Widgets")
        del master.MasterView.model_title

        # or it may specify model name
        master.MasterView.model_name = 'Blaster'
        self.assertEqual(master.MasterView.get_config_title(), "Blasters")
        del master.MasterView.model_name

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        master.MasterView.model_class = MyModel
        self.assertEqual(master.MasterView.get_config_title(), "Dinosaurs")
        del master.MasterView.model_class

    ##############################
    # support methods
    ##############################

    def test_get_index_title(self):
        master.MasterView.model_title_plural = "Wutta Widgets"
        view = master.MasterView(self.request)
        self.assertEqual(view.get_index_title(), "Wutta Widgets")
        del master.MasterView.model_title_plural

    def test_render_to_response(self):

        def widgets(request): return {}
        self.pyramid_config.add_route('widgets', '/widgets/')
        self.pyramid_config.add_view(widgets, route_name='widgets')

        # basic sanity check using /master/index.mako
        # (nb. it skips /widgets/index.mako since that doesn't exist)
        master.MasterView.model_name = 'Widget'
        view = master.MasterView(self.request)
        response = view.render_to_response('index', {})
        self.assertIsInstance(response, Response)
        del master.MasterView.model_name

        # basic sanity check using /appinfo/index.mako
        master.MasterView.model_name = 'AppInfo'
        master.MasterView.route_prefix = 'appinfo'
        master.MasterView.url_prefix = '/appinfo'
        view = master.MasterView(self.request)
        response = view.render_to_response('index', {})
        self.assertIsInstance(response, Response)
        del master.MasterView.model_name
        del master.MasterView.route_prefix
        del master.MasterView.url_prefix

        # bad template name causes error
        master.MasterView.model_name = 'Widget'
        self.assertRaises(IOError, view.render_to_response, 'nonexistent', {})
        del master.MasterView.model_name

    ##############################
    # view methods
    ##############################

    def test_index(self):
        
        # basic sanity check using /appinfo
        master.MasterView.model_name = 'AppInfo'
        master.MasterView.route_prefix = 'appinfo'
        master.MasterView.template_prefix = '/appinfo'
        view = master.MasterView(self.request)
        response = view.index()
        del master.MasterView.model_name
        del master.MasterView.route_prefix
        del master.MasterView.template_prefix

    def test_configure(self):
        model = self.app.model

        # setup
        master.MasterView.model_name = 'AppInfo'
        master.MasterView.route_prefix = 'appinfo'
        master.MasterView.template_prefix = '/appinfo'

        # mock settings
        settings = [
            {'name': 'wutta.app_title'},
            {'name': 'wutta.foo', 'value': 'bar'},
            {'name': 'wutta.flag', 'type': bool},
            {'name': 'wutta.number', 'type': int, 'default': 42},
            {'name': 'wutta.value1', 'save_if_empty': True},
            {'name': 'wutta.value2', 'save_if_empty': False},
        ]

        view = master.MasterView(self.request)
        with patch.object(self.request, 'current_route_url',
                          return_value='/appinfo/configure'):
            with patch.object(master.MasterView, 'configure_get_simple_settings',
                              return_value=settings):
                with patch.object(master, 'Session', return_value=self.session):

                    # get the form page
                    response = view.configure()
                    self.assertIsInstance(response, Response)

                    # post request to save settings
                    self.request.method = 'POST'
                    self.request.POST = {
                        'wutta.app_title': 'Wutta',
                        'wutta.foo': 'bar',
                        'wutta.flag': 'true',
                    }
                    response = view.configure()
                    # nb. should get redirect back to configure page
                    self.assertIsInstance(response, HTTPFound)

                    # should now have 5 settings
                    count = self.session.query(model.Setting).count()
                    self.assertEqual(count, 5)
                    get_setting = functools.partial(self.app.get_setting, self.session)
                    self.assertEqual(get_setting('wutta.app_title'), 'Wutta')
                    self.assertEqual(get_setting('wutta.foo'), 'bar')
                    self.assertEqual(get_setting('wutta.flag'), 'true')
                    self.assertEqual(get_setting('wutta.number'), '42')
                    self.assertEqual(get_setting('wutta.value1'), '')
                    self.assertEqual(get_setting('wutta.value2'), None)

                    # post request to remove settings
                    self.request.method = 'POST'
                    self.request.POST = {'remove_settings': '1'}
                    response = view.configure()
                    # nb. should get redirect back to configure page
                    self.assertIsInstance(response, HTTPFound)

                    # should now have 0 settings
                    count = self.session.query(model.Setting).count()
                    self.assertEqual(count, 0)

        # teardown
        del master.MasterView.model_name
        del master.MasterView.route_prefix
        del master.MasterView.template_prefix



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import patch, MagicMock

from fanstatic import Library, Resource
from pyramid import testing

from wuttjamaican.conf import WuttaConfig
from wuttaweb import util


class TestGetLibVer(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest()
        self.request.wutta_config = self.config

    def test_buefy_default(self):
        version = util.get_libver(self.request, 'buefy')
        self.assertEqual(version, 'latest')

    def test_buefy_custom_old(self):
        self.config.setdefault('wuttaweb.buefy_version', '0.9.29')
        version = util.get_libver(self.request, 'buefy')
        self.assertEqual(version, '0.9.29')

    def test_buefy_custom_new(self):
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = util.get_libver(self.request, 'buefy')
        self.assertEqual(version, '0.9.29')

    def test_buefy_configured_only(self):
        version = util.get_libver(self.request, 'buefy', configured_only=True)
        self.assertIsNone(version)

    def test_buefy_default_only(self):
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = util.get_libver(self.request, 'buefy', default_only=True)
        self.assertEqual(version, 'latest')

    def test_buefy_css_default(self):
        version = util.get_libver(self.request, 'buefy.css')
        self.assertEqual(version, 'latest')

    def test_buefy_css_custom_old(self):
        # nb. this uses same setting as buefy (js)
        self.config.setdefault('wuttaweb.buefy_version', '0.9.29')
        version = util.get_libver(self.request, 'buefy.css')
        self.assertEqual(version, '0.9.29')

    def test_buefy_css_custom_new(self):
        # nb. this uses same setting as buefy (js)
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = util.get_libver(self.request, 'buefy.css')
        self.assertEqual(version, '0.9.29')

    def test_buefy_css_configured_only(self):
        version = util.get_libver(self.request, 'buefy.css', configured_only=True)
        self.assertIsNone(version)

    def test_buefy_css_default_only(self):
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = util.get_libver(self.request, 'buefy.css', default_only=True)
        self.assertEqual(version, 'latest')

    def test_vue_default(self):
        version = util.get_libver(self.request, 'vue')
        self.assertEqual(version, '2.6.14')

    def test_vue_custom_old(self):
        self.config.setdefault('wuttaweb.vue_version', '3.4.31')
        version = util.get_libver(self.request, 'vue')
        self.assertEqual(version, '3.4.31')

    def test_vue_custom_new(self):
        self.config.setdefault('wuttaweb.libver.vue', '3.4.31')
        version = util.get_libver(self.request, 'vue')
        self.assertEqual(version, '3.4.31')

    def test_vue_configured_only(self):
        version = util.get_libver(self.request, 'vue', configured_only=True)
        self.assertIsNone(version)

    def test_vue_default_only(self):
        self.config.setdefault('wuttaweb.libver.vue', '3.4.31')
        version = util.get_libver(self.request, 'vue', default_only=True)
        self.assertEqual(version, '2.6.14')

    def test_vue_resource_default(self):
        version = util.get_libver(self.request, 'vue_resource')
        self.assertEqual(version, 'latest')

    def test_vue_resource_custom(self):
        self.config.setdefault('wuttaweb.libver.vue_resource', '1.5.3')
        version = util.get_libver(self.request, 'vue_resource')
        self.assertEqual(version, '1.5.3')

    def test_fontawesome_default(self):
        version = util.get_libver(self.request, 'fontawesome')
        self.assertEqual(version, '5.3.1')

    def test_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.libver.fontawesome', '5.6.3')
        version = util.get_libver(self.request, 'fontawesome')
        self.assertEqual(version, '5.6.3')

    def test_bb_vue_default(self):
        version = util.get_libver(self.request, 'bb_vue')
        self.assertEqual(version, '3.4.31')

    def test_bb_vue_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_vue', '3.4.30')
        version = util.get_libver(self.request, 'bb_vue')
        self.assertEqual(version, '3.4.30')

    def test_bb_oruga_default(self):
        version = util.get_libver(self.request, 'bb_oruga')
        self.assertEqual(version, '0.8.12')

    def test_bb_oruga_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_oruga', '0.8.11')
        version = util.get_libver(self.request, 'bb_oruga')
        self.assertEqual(version, '0.8.11')

    def test_bb_oruga_bulma_default(self):
        version = util.get_libver(self.request, 'bb_oruga_bulma')
        self.assertEqual(version, '0.3.0')
        version = util.get_libver(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(version, '0.3.0')

    def test_bb_oruga_bulma_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_oruga_bulma', '0.2.11')
        version = util.get_libver(self.request, 'bb_oruga_bulma')
        self.assertEqual(version, '0.2.11')

    def test_bb_fontawesome_svg_core_default(self):
        version = util.get_libver(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(version, '6.5.2')

    def test_bb_fontawesome_svg_core_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_fontawesome_svg_core', '6.5.1')
        version = util.get_libver(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(version, '6.5.1')

    def test_bb_free_solid_svg_icons_default(self):
        version = util.get_libver(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(version, '6.5.2')

    def test_bb_free_solid_svg_icons_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_free_solid_svg_icons', '6.5.1')
        version = util.get_libver(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(version, '6.5.1')

    def test_bb_vue_fontawesome_default(self):
        version = util.get_libver(self.request, 'bb_vue_fontawesome')
        self.assertEqual(version, '3.0.6')

    def test_bb_vue_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_vue_fontawesome', '3.0.8')
        version = util.get_libver(self.request, 'bb_vue_fontawesome')
        self.assertEqual(version, '3.0.8')


libcache = Library('testing', 'libcache')
vue_js = Resource(libcache, 'vue.js')
vue_resource_js = Resource(libcache, 'vue_resource.js')
buefy_js = Resource(libcache, 'buefy.js')
buefy_css = Resource(libcache, 'buefy.css')
fontawesome_js = Resource(libcache, 'fontawesome.js')
bb_vue_js = Resource(libcache, 'bb_vue.js')
bb_oruga_js = Resource(libcache, 'bb_oruga.js')
bb_oruga_bulma_js = Resource(libcache, 'bb_oruga_bulma.js')
bb_oruga_bulma_css = Resource(libcache, 'bb_oruga_bulma.css')
bb_fontawesome_svg_core_js = Resource(libcache, 'bb_fontawesome_svg_core.js')
bb_free_solid_svg_icons_js = Resource(libcache, 'bb_free_solid_svg_icons.js')
bb_vue_fontawesome_js = Resource(libcache, 'bb_vue_fontawesome.js')


class TestGetLibUrl(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config)
        self.pyramid_config = testing.setUp(request=self.request)

    def tearDown(self):
        testing.tearDown()

    def setup_fanstatic(self):
        self.pyramid_config.include('pyramid_fanstatic')
        self.config.setdefault('wuttaweb.static_libcache.module',
                               'tests.test_util')

        needed = MagicMock()
        needed.library_url = MagicMock(return_value='/fanstatic')
        self.request.environ['fanstatic.needed'] = needed
        self.request.script_name = '/wutta'

    def test_buefy_default(self):
        url = util.get_liburl(self.request, 'buefy')
        self.assertEqual(url, 'https://unpkg.com/buefy@latest/dist/buefy.min.js')

    def test_buefy_custom(self):
        self.config.setdefault('wuttaweb.liburl.buefy', '/lib/buefy.js')
        url = util.get_liburl(self.request, 'buefy')
        self.assertEqual(url, '/lib/buefy.js')

    def test_buefy_default_only(self):
        self.config.setdefault('wuttaweb.liburl.buefy', '/lib/buefy.js')
        url = util.get_liburl(self.request, 'buefy', default_only=True)
        self.assertEqual(url, 'https://unpkg.com/buefy@latest/dist/buefy.min.js')

    def test_buefy_configured_only(self):
        url = util.get_liburl(self.request, 'buefy', configured_only=True)
        self.assertIsNone(url)

    def test_buefy_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'buefy')
        self.assertEqual(url, '/wutta/fanstatic/buefy.js')

    def test_buefy_css_default(self):
        url = util.get_liburl(self.request, 'buefy.css')
        self.assertEqual(url, 'https://unpkg.com/buefy@latest/dist/buefy.min.css')

    def test_buefy_css_custom(self):
        self.config.setdefault('wuttaweb.liburl.buefy.css', '/lib/buefy.css')
        url = util.get_liburl(self.request, 'buefy.css')
        self.assertEqual(url, '/lib/buefy.css')

    def test_buefy_css_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'buefy.css')
        self.assertEqual(url, '/wutta/fanstatic/buefy.css')

    def test_vue_default(self):
        url = util.get_liburl(self.request, 'vue')
        self.assertEqual(url, 'https://unpkg.com/vue@2.6.14/dist/vue.min.js')

    def test_vue_custom(self):
        self.config.setdefault('wuttaweb.liburl.vue', '/lib/vue.js')
        url = util.get_liburl(self.request, 'vue')
        self.assertEqual(url, '/lib/vue.js')

    def test_vue_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'vue')
        self.assertEqual(url, '/wutta/fanstatic/vue.js')

    def test_vue_resource_default(self):
        url = util.get_liburl(self.request, 'vue_resource')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/vue-resource@latest')

    def test_vue_resource_custom(self):
        self.config.setdefault('wuttaweb.liburl.vue_resource', '/lib/vue-resource.js')
        url = util.get_liburl(self.request, 'vue_resource')
        self.assertEqual(url, '/lib/vue-resource.js')

    def test_vue_resource_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'vue_resource')
        self.assertEqual(url, '/wutta/fanstatic/vue_resource.js')

    def test_fontawesome_default(self):
        url = util.get_liburl(self.request, 'fontawesome')
        self.assertEqual(url, 'https://use.fontawesome.com/releases/v5.3.1/js/all.js')

    def test_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.liburl.fontawesome', '/lib/fontawesome.js')
        url = util.get_liburl(self.request, 'fontawesome')
        self.assertEqual(url, '/lib/fontawesome.js')

    def test_fontawesome_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'fontawesome')
        self.assertEqual(url, '/wutta/fanstatic/fontawesome.js')

    def test_bb_vue_default(self):
        url = util.get_liburl(self.request, 'bb_vue')
        self.assertEqual(url, 'https://unpkg.com/vue@3.4.31/dist/vue.esm-browser.prod.js')

    def test_bb_vue_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_vue', '/lib/vue.js')
        url = util.get_liburl(self.request, 'bb_vue')
        self.assertEqual(url, '/lib/vue.js')

    def test_bb_vue_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'bb_vue')
        self.assertEqual(url, '/wutta/fanstatic/bb_vue.js')

    def test_bb_oruga_default(self):
        url = util.get_liburl(self.request, 'bb_oruga')
        self.assertEqual(url, 'https://unpkg.com/@oruga-ui/oruga-next@0.8.12/dist/oruga.mjs')

    def test_bb_oruga_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_oruga', '/lib/oruga.js')
        url = util.get_liburl(self.request, 'bb_oruga')
        self.assertEqual(url, '/lib/oruga.js')

    def test_bb_oruga_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'bb_oruga')
        self.assertEqual(url, '/wutta/fanstatic/bb_oruga.js')

    def test_bb_oruga_bulma_default(self):
        url = util.get_liburl(self.request, 'bb_oruga_bulma')
        self.assertEqual(url, 'https://unpkg.com/@oruga-ui/theme-bulma@0.3.0/dist/bulma.mjs')

    def test_bb_oruga_bulma_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_oruga_bulma', '/lib/oruga_bulma.js')
        url = util.get_liburl(self.request, 'bb_oruga_bulma')
        self.assertEqual(url, '/lib/oruga_bulma.js')

    def test_bb_oruga_bulma_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'bb_oruga_bulma')
        self.assertEqual(url, '/wutta/fanstatic/bb_oruga_bulma.js')

    def test_bb_oruga_bulma_css_default(self):
        url = util.get_liburl(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(url, 'https://unpkg.com/@oruga-ui/theme-bulma@0.3.0/dist/bulma.css')

    def test_bb_oruga_bulma_css_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_oruga_bulma_css', '/lib/oruga-bulma.css')
        url = util.get_liburl(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(url, '/lib/oruga-bulma.css')

    def test_bb_oruga_bulma_css_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(url, '/wutta/fanstatic/bb_oruga_bulma.css')

    def test_bb_fontawesome_svg_core_default(self):
        url = util.get_liburl(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-svg-core@6.5.2/+esm')

    def test_bb_fontawesome_svg_core_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_fontawesome_svg_core', '/lib/fontawesome-svg-core.js')
        url = util.get_liburl(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(url, '/lib/fontawesome-svg-core.js')

    def test_bb_fontawesome_svg_core_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(url, '/wutta/fanstatic/bb_fontawesome_svg_core.js')

    def test_bb_free_solid_svg_icons_default(self):
        url = util.get_liburl(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/@fortawesome/free-solid-svg-icons@6.5.2/+esm')

    def test_bb_free_solid_svg_icons_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_free_solid_svg_icons', '/lib/free-solid-svg-icons.js')
        url = util.get_liburl(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(url, '/lib/free-solid-svg-icons.js')

    def test_bb_free_solid_svg_icons_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(url, '/wutta/fanstatic/bb_free_solid_svg_icons.js')

    def test_bb_vue_fontawesome_default(self):
        url = util.get_liburl(self.request, 'bb_vue_fontawesome')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/@fortawesome/vue-fontawesome@3.0.6/+esm')

    def test_bb_vue_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_vue_fontawesome', '/lib/vue-fontawesome.js')
        url = util.get_liburl(self.request, 'bb_vue_fontawesome')
        self.assertEqual(url, '/lib/vue-fontawesome.js')

    def test_bb_vue_fontawesome_fanstatic(self):
        self.setup_fanstatic()
        url = util.get_liburl(self.request, 'bb_vue_fontawesome')
        self.assertEqual(url, '/wutta/fanstatic/bb_vue_fontawesome.js')


class TestGetFormData(TestCase):

    def setUp(self):
        self.config = WuttaConfig()

    def make_request(self, **kwargs):
        kwargs.setdefault('wutta_config', self.config)
        kwargs.setdefault('POST', {'foo1': 'bar'})
        kwargs.setdefault('json_body', {'foo2': 'baz'})
        return testing.DummyRequest(**kwargs)

    def test_default(self):
        request = self.make_request()
        data = util.get_form_data(request)
        self.assertEqual(data, {'foo1': 'bar'})

    def test_is_xhr(self):
        request = self.make_request(POST=None, is_xhr=True)
        data = util.get_form_data(request)
        self.assertEqual(data, {'foo2': 'baz'})

    def test_content_type(self):
        request = self.make_request(POST=None, content_type='application/json')
        data = util.get_form_data(request)
        self.assertEqual(data, {'foo2': 'baz'})


class TestGetCsrfToken(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_same_token(self):

        # same token returned for same request
        # TODO: dummy request is always returning same token!
        # so this isn't really testing anything.. :(
        first = util.get_csrf_token(self.request)
        self.assertIsNotNone(first)
        second = util.get_csrf_token(self.request)
        self.assertEqual(first, second)

        # TODO: ideally would make a new request here and confirm it
        # gets a different token, but see note above..

    def test_new_token(self):

        # nb. dummy request always returns same token, so must
        # trick it into thinking it doesn't have one yet
        with patch.object(self.request.session, 'get_csrf_token', return_value=None):
            token = util.get_csrf_token(self.request)
            self.assertIsNotNone(token)


class TestRenderCsrfToken(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_basics(self):
        html = util.render_csrf_token(self.request)
        self.assertIn('type="hidden"', html)
        self.assertIn('name="_csrf"', html)
        token = util.get_csrf_token(self.request)
        self.assertIn(f'value="{token}"', html)



# -*- coding: utf-8; -*-

from unittest import TestCase

from wuttjamaican.testing import FileConfigTestCase

from pyramid.config import Configurator
from pyramid.router import Router

from wuttaweb import app as mod
from wuttjamaican.conf import WuttaConfig


class TestWebAppProvider(TestCase):

    def test_basic(self):
        # nb. just normal usage here, confirm it does the one thing we
        # need it to..
        config = WuttaConfig()
        app = config.get_app()
        handler = app.get_web_handler()


class TestMakeWuttaConfig(FileConfigTestCase):

    def test_config_path_required(self):

        # settings must define config path, else error
        settings = {}
        self.assertRaises(ValueError, mod.make_wutta_config, settings)

    def test_basic(self):

        # mock path to config file
        myconf = self.write_file('my.conf', '')
        settings = {'wutta.config': myconf}

        # can make a config okay
        config = mod.make_wutta_config(settings)

        # and that config is also stored in settings
        self.assertIn('wutta_config', settings)
        self.assertIs(settings['wutta_config'], config)


class TestMakePyramidConfig(TestCase):

    def test_basic(self):
        settings = {}
        config = mod.make_pyramid_config(settings)
        self.assertIsInstance(config, Configurator)


class TestMain(FileConfigTestCase):

    def test_basic(self):
        global_config = None
        myconf = self.write_file('my.conf', '')
        settings = {'wutta.config': myconf}
        app = mod.main(global_config, **settings)
        self.assertIsInstance(app, Router)



# -*- coding: utf-8; -*-

from unittest import TestCase

from pyramid import testing
from pyramid.httpexceptions import HTTPFound, HTTPForbidden, HTTPNotFound

from wuttjamaican.conf import WuttaConfig
from wuttaweb.views import base
from wuttaweb.forms import Form


class TestView(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.app = self.config.get_app()
        self.request = testing.DummyRequest(wutta_config=self.config)
        self.view = base.View(self.request)

    def test_basic(self):
        self.assertIs(self.view.request, self.request)
        self.assertIs(self.view.config, self.config)
        self.assertIs(self.view.app, self.app)

    def test_forbidden(self):
        error = self.view.forbidden()
        self.assertIsInstance(error, HTTPForbidden)

    def test_make_form(self):
        form = self.view.make_form()
        self.assertIsInstance(form, Form)

    def test_notfound(self):
        error = self.view.notfound()
        self.assertIsInstance(error, HTTPNotFound)

    def test_redirect(self):
        error = self.view.redirect('/')
        self.assertIsInstance(error, HTTPFound)
        self.assertEqual(error.location, '/')
